import logging

import mlflow

from databricks.rag_eval import context
from databricks.rag_eval.mlflow import mlflow_utils
from databricks.rag_eval.monitoring import assessments, entities
from databricks.rag_eval.monitoring import external as external_monitoring
from databricks.rag_eval.monitoring import utils as monitoring_utils
from databricks.rag_eval.utils import uc_utils

_logger = logging.getLogger(__name__)


def _warn_monitoring_config_deprecation():
    print(
        "WARNING: The `monitoring_config` argument is deprecated and will be removed in a future version. "
        "Please use the `assessments_config` argument instead.\n"
    )


def _get_managed_evals_client():
    return context.get_context().build_managed_evals_client()


def _parse_monitoring_config(
    monitoring_config: dict | entities.MonitoringConfig,
) -> entities.MonitoringConfig:
    assert monitoring_config is not None, "monitoring_config is required"
    monitoring_config = entities.MonitoringConfig.from_dict(monitoring_config)

    # Validate sampling.
    assert isinstance(
        monitoring_config.sample, (int, float)
    ), "monitoring_config.sample must be a number"
    assert (
        0 <= monitoring_config.sample <= 1
    ), "monitoring_config.sample must be between 0 and 1"

    # Validate paused.
    assert monitoring_config.paused is None or isinstance(
        monitoring_config.paused, bool
    ), "monitoring_config.paused must be a boolean"

    # Validate metrics.
    assert monitoring_config.metrics is None or (
        isinstance(monitoring_config.metrics, list)
    ), "monitoring_config.metrics must be a list of strings"

    # Validate guidelines
    if monitoring_config.global_guidelines is not None:
        error_msg = "monitoring_config.global_guidelines must be a dictionary of {<guideline_name>: [<guidelines>]}"
        assert isinstance(monitoring_config.global_guidelines, dict), error_msg
        assert all(
            isinstance(g, list) for g in monitoring_config.global_guidelines.values()
        ), error_msg
        assert all(
            isinstance(e, str)
            for g in monitoring_config.global_guidelines.values()
            for e in g
        ), error_msg
        assert all(
            isinstance(k, str) for k in monitoring_config.global_guidelines.keys()
        ), error_msg

    return monitoring_config


@context.eval_context
def create_monitor(
    endpoint_name: str,
    *,
    assessments_config: dict | assessments.AssessmentsSuiteConfig | None = None,
    experiment_id: str | None = None,
    monitoring_config: dict | entities.MonitoringConfig | None = None,
) -> entities.Monitor:
    """
    Create a monitor for a Databricks serving endpoint.

    Args:
        endpoint_name: The name of the serving endpoint.
        assessments_config: The configuration for the suite of assessments to be run on traces.
        experiment_id: The experiment ID to log the monitoring results. Defaults to the experiment
            used to log the model that is serving the provided `endpoint_name`.
        monitoring_config: Deprecated. The monitoring configuration.
    Returns:
        The monitor for the serving endpoint.
    """
    if (monitoring_config is None) == (assessments_config is None):
        raise ValueError(
            "Exactly one of `monitoring_config` or `assessments_config` must be specified."
        )

    if monitoring_config:
        _warn_monitoring_config_deprecation()
        monitoring_config = _parse_monitoring_config(monitoring_config)

        # don't let global_guidelines be set without guideline_adherence metric
        if (
            monitoring_config.global_guidelines
            and assessments.AI_JUDGE_GUIDELINE_ADHERENCE
            not in (monitoring_config.metrics or [])
        ):
            raise ValueError(
                "Global guidelines can only be set if 'guideline_adherence' is included in `metrics`."
            )

        assessments_config = monitoring_config.to_assessments_suite_config()

    if isinstance(assessments_config, dict):
        assessments_config = assessments.AssessmentsSuiteConfig.from_dict(
            assessments_config
        )

    monitor = _create_monitor_internal(endpoint_name, assessments_config, experiment_id)
    print(f'Created monitor for experiment "{monitor.experiment_id}".')
    print(f"\nView monitoring page: {monitor.monitoring_page_url}")

    if monitor.assessments_config.assessments:
        print("\nAssessments:")
        for assessment in monitor.assessments_config.assessments:
            if isinstance(assessment, assessments.BuiltinJudge):
                print(f"• {assessment.name}")
            elif isinstance(assessment, assessments.GuidelinesJudge):
                print(f"• {assessments.AI_JUDGE_GUIDELINE_ADHERENCE}")
    else:
        print(
            "\nNo assessments specified. To override the assessments, include `BuiltinJudge` or `GuidelinesJudge` instances in the `assessments` field in `assessments_config`."
        )

    return monitor


@context.eval_context
def _create_monitor_internal(
    endpoint_name: str,
    assessments_config: assessments.AssessmentsSuiteConfig,
    experiment_id: str | None = None,
) -> entities.Monitor:
    """
    Internal implementation of create_monitor. This function is called by both `create_monitor` and `agents.deploy`.
    """

    if experiment_id is None:
        # Infer the experiment ID and the workspace path from the current environment.
        experiment = mlflow_utils.infer_experiment_from_endpoint(endpoint_name)
    else:
        experiment = mlflow.get_experiment(experiment_id)

    return _get_managed_evals_client().create_monitor(
        endpoint_name=endpoint_name,
        assessments_config=assessments_config,
        experiment_id=experiment.experiment_id,
    )


@context.eval_context
def create_external_monitor(
    *,
    catalog_name: str,
    schema_name: str,
    assessments_config: assessments.AssessmentsSuiteConfig | dict,
    experiment_id: str | None = None,
    experiment_name: str | None = None,
) -> entities.ExternalMonitor:
    """Create a monitor for a GenAI application served outside Databricks.

    Args:
        catalog_name (str): The name of the catalog in UC to create the checkpoint table in.
        schema_name (str): The name of the schema in UC to create the checkpoint table in.
        assessments_config (AssessmentsSuiteConfig | dict): The configuration for the suite of
            assessments to be run on traces from the GenAI application.
        experiment_id (str | None, optional): ID of Mlflow experiment
            that the monitor should be associated with. Defaults to the
            currently active experiment.
        experiment_name (str | None, optional): The name of the Mlflow experiment that the monitor
            should be associated with. Defaults to the currently active experiment.

    Returns:
        ExternalMonitor: The created monitor.
    """
    if not catalog_name or not schema_name:
        raise ValueError(
            "Both catalog_name and schema_name must be provided and non-empty."
        )

    # fresh assessments suite config validation and defaults
    if assessments_config is None:
        raise ValueError("assessments_config is required")
    if isinstance(assessments_config, dict):
        assessments_config = assessments.AssessmentsSuiteConfig.from_dict(
            assessments_config
        )
    if assessments_config.sample is None:
        assessments_config.sample = 1.0
    experiment = monitoring_utils.get_databricks_mlflow_experiment(
        experiment_id=experiment_id,
        experiment_name=experiment_name,
    )
    ckpt_table_entity_name = monitoring_utils.create_checkpoint_table_entity_name(
        experiment_name=experiment.name,
    )
    ckpt_table_uc_entity = uc_utils.UnityCatalogEntity.from_fullname(
        f"{catalog_name}.{schema_name}.{ckpt_table_entity_name}"
    )
    endpoint_name = external_monitoring.build_endpoint_name(ckpt_table_uc_entity.entity)

    monitor: entities.ExternalMonitor = _get_managed_evals_client().create_monitor(
        endpoint_name=endpoint_name,
        assessments_config=assessments_config,
        experiment_id=experiment.experiment_id,
        monitoring_table=ckpt_table_uc_entity.fullname,
    )

    # Create the monitoring table now so that it exists immediately after this function finishes,
    # but allows the service to perform validations.
    try:
        external_monitoring.create_monitoring_table(ckpt_table_uc_entity)
    except Exception as e:
        # Cleanup monitor if monitoring table creation fails, since
        # monitoring job will always fail to run.
        _logger.debug(
            f"Failed to create checkpoint table '{ckpt_table_uc_entity.fullname}': {e}. Deleting monitor to cleanup..."
        )
        _get_managed_evals_client().delete_monitor(endpoint_name=endpoint_name)
        raise e

    print(f'Created monitor for experiment "{monitor.experiment_id}".')
    print(f"\nView monitoring page: {monitor.monitoring_page_url}")

    if monitor.assessments_config.assessments:
        print("\nBuilt-in AI judges:")
        for assessment in monitor.assessments_config.assessments:
            if isinstance(assessment, assessments.BuiltinJudge):
                metric = str(assessment.name)
                print(f"• {metric}")
            elif isinstance(assessment, assessments.GuidelinesJudge):
                print(f"• {assessments.AI_JUDGE_GUIDELINE_ADHERENCE}")
    else:
        print(
            "\nNo built-in AI judges specified. To add built-in AI judges, include `BuiltinJudge` instances in the `assessments` field in `assessments_config`."
        )

    print(
        f"\nCreated external monitor successfully using experiment_id '{monitor.experiment_id}'.\n\n"
        "You will need this experiment_id when you set your agent's tracing destination to "
        f"`Databricks(experiment_id={monitor.experiment_id})`"
    )
    return monitor


@context.eval_context
def get_monitor(
    *,
    endpoint_name: str,
) -> entities.Monitor:
    """
    Retrieves a monitor for a Databricks serving endpoint.

    Args:
        endpoint_name (str, optional): The name of the agent's serving endpoint.

    Returns:
        Monitor | ExternalMonitor metadata. For external monitors, this will include the status of the ingestion endpoint.
    """
    monitor = _get_monitor_internal(
        endpoint_name=endpoint_name,
    )
    print("Monitor URL: ", monitor.monitoring_page_url)
    return monitor


@context.eval_context
def _get_monitor_internal(
    *,
    endpoint_name: str | None = None,
) -> entities.Monitor:
    """
    Internal implementation of get_monitor.
    """
    return _get_managed_evals_client().get_monitor(
        endpoint_name=endpoint_name,
    )


@context.eval_context
def get_external_monitor(
    *,
    experiment_id: str | None = None,
    experiment_name: str | None = None,
) -> entities.ExternalMonitor:
    """Gets the monitor for a GenAI application served outside Databricks.

    Args:
        experiment_id (str | None, optional): ID of the Mlflow experiment that
            the monitor is associated with. Defaults to None.
        experiment_name (str | None, optional): Name of the Mlflow experiment that
            the monitor is associated with. Defaults to None.

    Raises:
        ValueError: When neither experiment_id nor experiment_name is provided.
        ValueError: When no monitor is found for the given experiment_id or experiment_name.

    Returns:
        entities.ExternalMonitor: The retrieved external monitor.
    """
    if experiment_id is None and experiment_name is None:
        raise ValueError("Please provide either an experiment_id or experiment_name.")

    experiment = monitoring_utils.get_databricks_mlflow_experiment(
        experiment_id=experiment_id,
        experiment_name=experiment_name,
    )
    monitors: list[
        entities.ExternalMonitor
    ] = _get_managed_evals_client().list_monitors(
        experiment_id=experiment.experiment_id,
    )
    if not monitors:
        if experiment_name:
            raise ValueError(
                f"No monitor found for experiment_name '{experiment_name}'."
            )
        else:
            raise ValueError(f"No monitor found for experiment_id '{experiment_id}'.")

    return _get_managed_evals_client().get_monitor(
        # There should be only one monitor per experiment
        monitoring_table=monitors[0]._checkpoint_table,
    )


@context.eval_context
def update_monitor(
    *,
    endpoint_name: str,
    assessments_config: dict | assessments.AssessmentsSuiteConfig | None = None,
    monitoring_config: dict | entities.MonitoringConfig | None = None,
) -> entities.Monitor:
    """
    Partially update a monitor for a serving endpoint.

    Args:
        endpoint_name (str, optional): The name of the agent's serving endpoint.
            Only supported for agents served on Databricks.
        assessments_config: The updated configuration for the suite of assessments to be run on traces.
            Partial updates of arrays is not supported, so assessments specified here will override your
            monitor’s assessments. If unspecified, non-nested fields like `sample` will not be updated.
        monitoring_config: Deprecated. The configuration change, using upsert semantics.

    Returns:
        Monitor: The updated monitor for the serving endpoint.
    """
    if (monitoring_config is None) == (assessments_config is None):
        raise ValueError(
            "Exactly one of `monitoring_config` or `assessments_config` must be specified."
        )

    if monitoring_config:
        _warn_monitoring_config_deprecation()
        if isinstance(monitoring_config, dict):
            monitoring_config = entities.MonitoringConfig.from_dict(monitoring_config)
        # don't let global_guidelines be set without guideline_adherence metric
        if (
            monitoring_config.global_guidelines
            and assessments.AI_JUDGE_GUIDELINE_ADHERENCE
            not in (monitoring_config.metrics or [])
        ):
            raise ValueError(
                "Global guidelines can only be set if 'guideline_adherence' is included in `metrics`."
            )
        assessments_config = monitoring_config.to_assessments_suite_config()

    if isinstance(assessments_config, dict):
        assessments_config = assessments.AssessmentsSuiteConfig.from_dict(
            assessments_config
        )

    return _get_managed_evals_client().update_monitor(
        endpoint_name=endpoint_name,
        assessments_config=assessments_config,
    )


@context.eval_context
def update_external_monitor(
    *,
    experiment_id: str | None = None,
    experiment_name: str | None = None,
    assessments_config: assessments.AssessmentsSuiteConfig | dict,
) -> entities.ExternalMonitor:
    """Updates the monitor for an GenAI application served outside Databricks.

    Args:
        assessments_config (assessments.AssessmentsSuiteConfig): The updated configuration for
            the suite of assessments to be run on traces from the AI system. Partial updates
            of arrays is not supported, so assessments specified here will override your monitor's
            assessments. If unspecified, non-nested fields like `sample` will not be updated.
        experiment_id (str | None, optional): ID of the Mlflow experiment that the monitor
            is associated with. Defaults to None.
        experiment_name (str | None, optional): Name of the Mlflow experiment that the monitor
            is associated with. Defaults to None.

    Raises:
        ValueError: When assessments_config is not provided.

    Returns:
        entities.ExternalMonitor: The updated external monitor.
    """

    if assessments_config is None:
        raise ValueError("assessments_config is required")
    if isinstance(assessments_config, dict):
        assessments_config = assessments.AssessmentsSuiteConfig.from_dict(
            assessments_config
        )

    monitor = get_external_monitor(
        experiment_id=experiment_id,
        experiment_name=experiment_name,
    )
    return _get_managed_evals_client().update_monitor(
        endpoint_name=monitor._legacy_ingestion_endpoint_name,
        assessments_config=assessments_config,
    )


@context.eval_context
def delete_monitor(
    *,
    endpoint_name: str | None = None,
) -> None:
    """
    Deletes a monitor for a Databricks serving endpoint.

    Args:
        endpoint_name (str, optional): The name of the agent's serving endpoint.
    """
    # for external monitors, find the endpoint name using the monitoring table
    return _get_managed_evals_client().delete_monitor(endpoint_name=endpoint_name)


@context.eval_context
def delete_external_monitor(
    *,
    experiment_id: str | None = None,
    experiment_name: str | None = None,
) -> None:
    """Deletes the monitor for a GenAI application served outside Databricks.

    Args:
        experiment_id (str | None, optional): ID of the Mlflow experiment that the monitor
            is associated with. Defaults to None.
        experiment_name (str | None, optional): Name of the Mlflow experiment that the monitor
            is associated with. Defaults to None.
    """

    monitor = get_external_monitor(
        experiment_id=experiment_id,
        experiment_name=experiment_name,
    )
    _get_managed_evals_client().delete_monitor(
        endpoint_name=monitor._legacy_ingestion_endpoint_name,
    )
